import { program } from "commander";
import { booleanPrompt, endPrompts, startPrompts, textPrompt } from "../utils/prompts.js";
import { addPropertyToWorkspaceOption } from "../utils/workspace.js";

/**
 * @typedef BooleanDefaultValue
 * @property {"boolean"} type
 * @property {boolean} default
 */

/**
 * @typedef StringDefaultValue
 * @property {"string"} type
 * @property {string} default
 */

/**
 * @typedef NumberDefaultValue
 * @property {"number"} type
 * @property {number} default
 */

/**
 * @typedef BaseQuestion
 * @property {string} question,
 */

/**
 * @typedef {BaseQuestion & BooleanDefaultValue} BooleanQuestion
 * @typedef {BaseQuestion & StringDefaultValue} StringQuestion
 * @typedef {BaseQuestion & NumberDefaultValue} NumberQuestion
 * @typedef {BooleanQuestion | StringQuestion | NumberQuestion} Question
 */

/**
 * @template {import("./config").ArgType} Args
 * @param {import("./config").ComposerConfig<Args>} config
 */
export function prepareAndParseCliOptions(config) {
    program.option("--path <string>", "Path to working directory");

    if (config.options) {
        for (const optionKey of Object.keys(config.options)) {
            const option = config.options[optionKey];

            program.option(`--${optionKey} [value]`, option.question);
        }
    }

    program.parse();
    const options = program.opts();
    return options;
}

/**
 *
 * @template {import("./config").ArgType} Args
 * @param {import("./config").ComposerConfig<Args>} config
 * @param {import("./config").Workspace<Args>} workspace
 * @param {import("commander").OptionValues} cliOptions
 * @returns {Promise}
 */
export async function askQuestionsAndAssignValuesToWorkspace(config, workspace, cliOptions) {
    if (!config.options) return;

    let needsToAskQuestions = false;
    for (const optionKey of Object.keys(config.options)) {
        const cliOption = cliOptions[optionKey];
        if (cliOption === undefined) {
            needsToAskQuestions = true;
        }
    }

    if (needsToAskQuestions) {
        startPrompts(config.metadata.package);
    }

    for (const optionKey of Object.keys(config.options)) {
        const cliOption = cliOptions[optionKey];
        if (cliOption !== undefined) {
            addPropertyToWorkspaceOption(workspace, optionKey, cliOption);

            continue;
        }

        const option = config.options[optionKey];
        let optionValue;

        if (option.type == "number" || option.type == "string") {
            optionValue = await textPrompt(option.question, "Not sure", "" + option.default);
        } else if (option.type == "boolean") {
            optionValue = await booleanPrompt(option.question, /** @type {boolean} */ (option.default));
        }

        addPropertyToWorkspaceOption(workspace, optionKey, optionValue);
    }

    if (needsToAskQuestions) {
        endPrompts(`You're all set!`);
    }
}

/**
 * @template {import("./config").ArgType} Args
 * @param {import("./config").ComposerConfig<Args>} config
 * @param {import("./config").Workspace<Args>} workspace
 * @returns
 */
export function ensureCorrectOptionTypes(config, workspace) {
    if (!config.options) {
        return;
    }

    let foundInvalidType = false;

    for (const optionKey of Object.keys(config.options)) {
        const option = config.options[optionKey];
        const value = workspace.options[optionKey];

        if (option.type == "boolean" && typeof value == "boolean") {
            continue;
        } else if (option.type == "number" && typeof value == "number") {
            continue;
        } else if (
            option.type == "number" &&
            typeof value == "string" &&
            typeof parseInt(value) == "number" &&
            !isNaN(parseInt(value))
        ) {
            addPropertyToWorkspaceOption(workspace, optionKey, parseInt(value));
            continue;
        } else if (option.type == "string" && typeof value == "string") {
            continue;
        }

        foundInvalidType = true;
        console.log(`Option ${optionKey} needs to be of type ${option.type} but was of type ${typeof value}!`);
    }

    if (foundInvalidType) {
        process.exit(0);
    }
}
