import { getCssAstEditor, getHtmlAstEditor, getJsAstEditor } from "@svelte-compose/ast-manipulation";
import {
    parseHtml,
    parsePostcss,
    parseScript,
    parseSvelteFile,
    serializeHtml,
    serializePostcss,
    serializeScript,
    serializeSvelteFile,
} from "@svelte-compose/ast-tooling";
import { fileExistsWorkspace, format, readFile, writeFile } from "./utils.js";

/**
 * @typedef {import("../composer/config").ArgType} ArgType
 */

/**
 * @template {ArgType} Args
 * @typedef {import("../utils/workspace").Workspace<Args>} Workspace
 */

/**
 * @template {ArgType} Args
 * @typedef {import("../composer/config").ConditionDefinition<Args>} ConditionDefinition
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").JsAstEditor & Workspace<Args>} ScriptFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef  ScriptFileType
 * @property {"script"} contentType
 * @property {(editor: ScriptFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {ScriptFileType<Args> & BaseFile<Args>} ScriptFile
 */

/**
 * @template {ArgType} Args
 * @typedef {{ content: string } & Workspace<Args>} TextFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef TextFileType
 * @property {"text"} contentType
 * @property {(editor: TextFileEditorArgs<Args>) => string} content
 */

/**
 * @template {ArgType} Args
 * @typedef {TextFileType<Args> & BaseFile<Args>} TextFile
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").SvelteAstEditor & Workspace<Args>} SvelteFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef SvelteFileType
 * @property {"svelte"} contentType
 * @property {(editor: SvelteFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {SvelteFileType<Args> & BaseFile<Args>} SvelteFile
 */

/**
 * @template {ArgType} Args
 * @typedef {{ data: any} & Workspace<Args>} JsonFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef JsonFileType
 * @property {"json"} contentType
 * @property {(editor: JsonFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {JsonFileType<Args> & BaseFile<Args>} JsonFile
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").HtmlAstEditor & Workspace<Args>} HtmlFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef  HtmlFileType
 * @property {"html"} contentType
 * @property {(editor: HtmlFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {HtmlFileType<Args> & BaseFile<Args>} HtmlFile
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").CssAstEditor & Workspace<Args>} CssFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef  CssFileType
 * @property {"css"} contentType
 * @property {(editor: CssFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {CssFileType<Args> & BaseFile<Args>} CssFile
 */

/**
 * @template {ArgType} Args
 * @typedef BaseFile
 * @property {(options: Workspace<Args>) => string} name
 * @property {ConditionDefinition<Args>} [condition]
 */

/**
 * @template {ArgType} Args
 * @param {import("../composer/config").FileTypes<Args>[]} files
 * @param {Workspace<Args>} workspace
 */
export async function createOrUpdateFiles(files, workspace) {
    for (const fileDetails of files) {
        if (fileDetails.condition && !fileDetails.condition(workspace)) {
            continue;
        }

        const exists = await fileExistsWorkspace(workspace, fileDetails.name(workspace));

        let content = "";
        if (!exists) {
            content = "";
        } else {
            content = await readFile(workspace, fileDetails.name(workspace));
        }

        if (fileDetails.contentType == "script") {
            content = handleScriptFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "text") {
            content = handleTextFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "svelte") {
            content = handleSvelteFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "json") {
            content = handleJsonFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "css") {
            content = handleCssFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "html") {
            content = handleHtmlFile(content, fileDetails, workspace);
        }

        content = await format(workspace, fileDetails.name(workspace), content);
        await writeFile(workspace, fileDetails.name(workspace), content);
    }
}

/**
 *
 * @template {ArgType} Args
 * @param {string} content
 * @param {HtmlFileType<Args>} fileDetails
 * @param {Workspace<Args>} workspace
 * @returns {string}
 */
function handleHtmlFile(content, fileDetails, workspace) {
    const ast = parseHtml(content);
    fileDetails.content({ ...getHtmlAstEditor(ast), ...workspace });
    content = serializeHtml(ast);
    return content;
}

/**
 *
 * @template {ArgType} Args
 * @param {string} content
 * @param {CssFileType<Args>} fileDetails
 * @param {Workspace<Args>} workspace
 * @returns {string}
 */
function handleCssFile(content, fileDetails, workspace) {
    const ast = parsePostcss(content);
    fileDetails.content({ ...getCssAstEditor(ast), ...workspace });
    content = serializePostcss(ast);
    return content;
}

/**
 *
 * @template {ArgType} Args
 * @param {string} content
 * @param {JsonFileType<Args>} fileDetails
 * @param {Workspace<Args>} workspace
 * @returns {string}
 */
function handleJsonFile(content, fileDetails, workspace) {
    const data = JSON.parse(content);
    fileDetails.content({ data, ...workspace });
    content = JSON.stringify(data, null, 4);
    return content;
}

/**
 *
 * @template {ArgType} Args
 * @param {string} content
 * @param {SvelteFileType<Args>} fileDetails
 * @param {Workspace<Args>} workspace
 * @returns {string}
 */
function handleSvelteFile(content, fileDetails, workspace) {
    const { jsAst, htmlAst, cssAst } = parseSvelteFile(content);

    fileDetails.content({
        js: getJsAstEditor(jsAst),
        html: getHtmlAstEditor(htmlAst),
        css: getCssAstEditor(cssAst),
        ...workspace,
    });

    return serializeSvelteFile({ jsAst, htmlAst, cssAst });
}

/**
 *
 * @template {ArgType} Args
 * @param {string} content
 * @param {TextFileType<Args>} fileDetails
 * @param {Workspace<Args>} workspace
 * @returns {string}
 */
function handleTextFile(content, fileDetails, workspace) {
    content = fileDetails.content({ content, ...workspace });
    return content;
}

/**
 *
 * @template {ArgType} Args
 * @param {string} content
 * @param {ScriptFileType<Args>} fileDetails
 * @param {Workspace<Args>} workspace
 * @returns {string}
 */
function handleScriptFile(content, fileDetails, workspace) {
    const ast = parseScript(content);

    fileDetails.content({
        ...getJsAstEditor(ast),
        ...workspace,
    });
    content = serializeScript(ast);
    return content;
}
