import fs from "fs/promises";
import path from "node:path";
import prettier from "prettier";

/**
 *
 * @template {import("../composer/config.js").ArgType} Args
 * @param {import("../composer/config.js").Workspace<Args>} workspace
 * @param {string} filePath
 * @returns
 */
export async function readFile(workspace, filePath) {
    const fullFilePath = getFilePath(workspace.cwd, filePath);

    if (!(await fileExistsWorkspace(workspace, filePath))) {
        return "";
    }

    const buffer = await fs.readFile(fullFilePath);
    const text = buffer.toString();

    return text;
}

/**
 *
 * @template {import("../composer/config.js").ArgType} Args
 * @param {import("../composer/config.js").Workspace<Args>} workspace
 * @param {string} filePath
 * @param {string} content
 */
export async function writeFile(workspace, filePath, content) {
    const fullFilePath = getFilePath(workspace.cwd, filePath);
    const fullDirectoryPath = path.dirname(fullFilePath);

    if (!(await directoryExists(fullDirectoryPath))) {
        await fs.mkdir(fullDirectoryPath);
    }

    await fs.writeFile(fullFilePath, content);
}

/**
 *
 * @template {import("../composer/config.js").ArgType} Args
 * @param {import("../composer/config.js").Workspace<Args>} workspace
 * @param {string} filePath
 * @returns {Promise<boolean>}
 */
export async function fileExistsWorkspace(workspace, filePath) {
    const fullFilePath = getFilePath(workspace.cwd, filePath);
    return await fileExists(fullFilePath);
}

/**
 *
 * @param {string} filePath
 * @returns {Promise<boolean>}
 */
export async function fileExists(filePath) {
    try {
        await fs.access(filePath, fs.constants.F_OK);
        return true;
    } catch (error) {
        return false;
    }
}

/**
 *
 * @param {string} directoryPath
 * @returns {Promise<boolean>}
 */
export async function directoryExists(directoryPath) {
    return await fileExists(directoryPath);
}

/**
 *
 * @param {string} cwd
 * @param {string} fileName
 * @returns {string}
 */
export function getFilePath(cwd, fileName) {
    return path.join(cwd, fileName);
}

/**
 * @template {import("../composer/config.js").ArgType} Args
 * @param {import("../composer/config.js").Workspace<Args>} workspace
 * @param {string} path
 * @param {string} content
 * @returns {Promise<string>}
 */
export async function format(workspace, path, content) {
    const fullPath = getFilePath(workspace.cwd, path);
    return await formatFileWithPrettier(fullPath, content);
}

/**
 *
 * @param {string} path
 * @param {string} content
 * @returns {Promise<string>}
 */
async function formatFileWithPrettier(path, content) {
    const options = await prettier.resolveConfig(path);

    try {
        return await prettier.format(content, {
            ...options,
            filepath: path,
            plugins: ["prettier-plugin-svelte"],
        });
    } catch (error) {
        // for some untypical file extensions prettier fails because
        // it's unable to find the appropriate parse. As this has only
        // to do with formatting and not functionality, we can
        // just skip formatting in this case.
        return content;
    }
}

export const commonFilePaths = {
    packageJsonFilePath: "package.json",
    svelteConfigFilePath: "svelte.config.js",
};
