import { spawnSync } from "child_process";
import * as path from "path";
import { booleanPrompt, endPrompts, selectPrompt, startPrompts, textPrompt } from "./prompts.js";
import { commonFilePaths, directoryExists, fileExists } from "../files/utils.js";
import { getPackageJson } from "./common.js";
import { createEmptyWorkspace } from "./workspace.js";

/**
 *
 * @param {string} cwd
 * @returns {Promise<string>}
 */
export async function detectOrCreateProject(cwd) {
    let workingDirectory = await detectSvelteDirectory(cwd);
    if (!workingDirectory) {
        console.log("Please create a new project first");

        const { projectCreated, directory } = await createProject(cwd);

        if (!projectCreated) {
            console.log("Template initializer failed, please see output above.");
            process.exit(0);
        }

        console.clear();
        workingDirectory = directory;
    }

    return workingDirectory;
}

/**
 *
 * @param {string} directoryPath
 * @returns {Promise<string | null>}
 */
export async function detectSvelteDirectory(directoryPath) {
    if (!directoryPath) return null;

    const packageJsonPath = path.join(directoryPath, commonFilePaths.packageJsonFilePath);
    const parentDirectoryPath = path.normalize(path.join(directoryPath, ".."));
    const isRoot = parentDirectoryPath == directoryPath;

    if (!isRoot && !(await directoryExists(directoryPath))) {
        return await detectSvelteDirectory(parentDirectoryPath);
    }

    if (!isRoot && !(await fileExists(packageJsonPath))) {
        return await detectSvelteDirectory(parentDirectoryPath);
    }

    if (isRoot && !(await fileExists(packageJsonPath))) {
        return null;
    }

    const emptyWorkspace = createEmptyWorkspace();
    emptyWorkspace.cwd = directoryPath;
    const packageJson = await getPackageJson(emptyWorkspace);

    if ("svelte" in packageJson.devDependencies) {
        return directoryPath;
    } else if (!isRoot) {
        return await detectSvelteDirectory(parentDirectoryPath);
    }

    return null;
}

/**
 *
 * @param {string} cwd
 * @returns {Promise<{projectCreated: boolean, directory: string}>}
 */
export async function createProject(cwd) {
    startPrompts("Create new project");

    const createNewProject = await booleanPrompt("Create new Project?", true);
    if (!createNewProject) {
        process.exit(0);
    }

    let relativePath = path.relative(process.cwd(), cwd);
    if (!relativePath) {
        relativePath = "./";
    }

    let directory = await textPrompt("Where should we create your project?", `  (hit Enter to use '${relativePath}')`);
    if (!directory) {
        directory = relativePath;
    }

    const projectType = await selectPrompt("Which project type do you want to create?", "kit", [
        { label: "SvelteKit", value: "kit" },
        { label: "Svelte", value: "svelte" },
    ]);

    let language = "js";
    if (projectType == "svelte") {
        language = /** @type {string} */ (
            await selectPrompt("Choose language", language, [
                { label: "JavaScript", value: "js" },
                { label: "TypeScript", value: "ts" },
            ])
        );
    }

    endPrompts("Initializing template...");

    let args = [];
    if (projectType == "kit") {
        args = ["init", "svelte@latest", directory];
    } else {
        const template = language == "ts" ? "svelte-ts" : "svelte";
        args = ["init", "vite@latest", directory, "--", "--template", template];
    }

    const program = spawnSync("npm", args, { stdio: "inherit", shell: true });

    if (program.status == 1) {
        console.log("cancelled or failed");
        return { projectCreated: false, directory: "" };
    }

    return { projectCreated: true, directory: path.join(process.cwd(), directory) };
}
