import { parseScript } from "@svelte-compose/ast-tooling";
import { getPackageJson } from "./common.js";
import { commonFilePaths, readFile } from "../files/utils.js";
import { getJsAstEditor } from "@svelte-compose/ast-manipulation";

/**
 * @typedef PrettierData
 * @property {boolean} installed
 * @property {unknown} config
 */

/**
 * @typedef TypescriptData
 * @property {boolean} installed
 * @property {unknown} config
 */

/**
 * @typedef SvelteKitData
 * @property {boolean} installed
 * @property {string} libDirectory
 * @property {string} routesDirectory
 */

/**
 * @template {import("../composer/config").ArgType} Args
 * @typedef Workspace
 * @property {import("../composer/config").ArgValues<Args>} options
 * @property {string} cwd
 * @property {PrettierData} prettier
 * @property {TypescriptData} typescript
 * @property {SvelteKitData} kit
 */

/**
 * @template {import("../composer/config").ArgType} Args
 * @returns {import("../composer/config").Workspace<Args>}
 */
export function createEmptyWorkspace() {
    return {
        // @ts-expect-error unable to make this typesafe
        options: {},
        cwd: "",
        prettier: {
            installed: false,
            config: "",
        },
        typescript: {
            installed: false,
            config: "",
        },
        kit: {
            installed: false,
            routesDirectory: "src/routes",
            libDirectory: "src/lib",
        },
    };
}

/**
 * @template {import("../composer/config").ArgType} Args
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @param {string} optionKey
 * @param {unknown} value
 */
export function addPropertyToWorkspaceOption(workspace, optionKey, value) {
    if (value === "true") {
        value = true;
    }

    if (value === "false") {
        value = false;
    }

    Object.defineProperty(workspace.options, optionKey, {
        value,
        writable: true,
        enumerable: true,
        configurable: true,
    });
}

/**
 * @template {import("../composer/config").ArgType} Args
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @param {string} workingDirectory
 */
export async function populateWorkspaceDetails(workspace, workingDirectory) {
    workspace.cwd = workingDirectory;

    const packageJson = await getPackageJson(workspace);
    workspace.typescript.installed = "tslib" in packageJson.devDependencies;
    workspace.prettier.installed = "prettier" in packageJson.devDependencies;
    workspace.kit.installed = "@sveltejs/kit" in packageJson.devDependencies;

    await parseSvelteConfigIntoWorkspace(workspace);
}

/**
 *
 * @template {import("../composer/config").ArgType} Args
 * @param {import("../composer/config").Workspace<Args>} workspace
 */
export async function parseSvelteConfigIntoWorkspace(workspace) {
    if (!workspace.kit.installed) return;
    const configText = await readFile(workspace, commonFilePaths.svelteConfigFilePath);
    const ast = parseScript(configText);
    const editor = await getJsAstEditor(ast);
    const variableDeclaration = /** @type {import("@svelte-compose/ast-tooling").VariableDeclaration} */ (
        ast.body.find((x) => x.type == "VariableDeclaration")
    );
    const variableDeclarator = /** @type {import("@svelte-compose/ast-tooling").VariableDeclarator} */ (
        variableDeclaration.declarations[0]
    );
    const objectExpression = /** @type {import("@svelte-compose/ast-tooling").ObjectExpression} */ (variableDeclarator.init);

    const kit = editor.object.property(objectExpression, "kit", editor.object.createEmpty());
    const files = editor.object.property(kit, "files", editor.object.createEmpty());
    const routes = editor.object.property(files, "routes", editor.common.createLiteral());
    const lib = editor.object.property(files, "lib", editor.common.createLiteral());

    if (routes.value) {
        workspace.kit.routesDirectory = /** @type {string} */ (routes.value);
    }
    if (lib.value) {
        workspace.kit.libDirectory = /** @type {string} */ (lib.value);
    }
}
